/*
 * This file is part of fkmines.
 * 
 * fkmines is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * fkmines is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
 * 
 * Copyright 2009, Richard Molitor
 */

#include <curses.h>
#include <stdlib.h>
#include <stdbool.h>
#include <error.h>
#include <time.h>
#include <string.h>

#include "generate.h"

#define MAJOR (0)
#define MINOR (2)

#define DEF (10)
#define MAXMSG (100)

static char closed = CLOSED_SYM;
static short *field;
static int lines = DEF, cols = DEF, mines = DEF, curx = 0, cury = 0;
static bool color;

void init (int argc, char *argv[]);
void redraw (int upper, int left, bool open);
void status (const char *string);
int open (int x, int y);
int mark (int x, int y);
void usage (char *progname);
void version (void);

enum extra_colors { COLOR_DARKGREEN=8, COLOR_DARKBLUE, COLOR_DARKRED,
  COLOR_DARKCYAN, COLOR_DARKGREY, COLOR_LIGHTGREY };

int main (int argc, char *argv[])
{
  bool quit = false,
       respawn = true;
  char msg[MAXMSG];
  int input,
      whites,
      opened = 0,
      marked = 0,
      topline = 0,
      leftcol = 0;
  float percentage = 0;
  long starttime;

  init (argc, argv);
  whites = countwhites (field);
  redraw (topline, leftcol, false);
  starttime = time (0);

  while (!quit)
  {
    input = getch ();
    switch (input)
    {
      case ERR:
        break;
      case KEY_LEFT:
      case 'a':
      case 'h':
        if (curx)
        {
          curx -= 2;
          move (cury, curx);
        }
        else if (leftcol)
          leftcol -= 2;
        break;
      case KEY_DOWN:
      case 's':
      case 'j':
        if (topline+cury < lines-1)
        {
          if (cury < LINES-2)
            move (++cury, curx);
          else topline++;
        }
        break;
      case KEY_RIGHT:
      case 'd':
      case 'l':
        if (leftcol+curx < 2*(cols-1))
        {
          if (curx < COLS-2)
          {
            curx += 2;
            move (cury, curx);
          }
          else leftcol += 2;
        }
        break;
      case KEY_UP:
      case 'w':
      case 'k':
        if (cury)
          move (--cury, curx);
        else if (topline)
          topline--;
        break;
      case 'x':
      case '!':
      case 'm':
        marked += mark ((curx+leftcol)/2, topline+cury);
        break;
      case ' ':
      case 'z':
      case 10:
      case KEY_ENTER:
        input = open ((curx+leftcol)/2, topline+cury);
        if (input > 0)
        {
          opened += input;
          redraw (topline, leftcol, false);
        }
        else if (input == -1)
        {
          redraw (topline, leftcol, true);
          snprintf (msg, MAXMSG,
                "You lost after %li seconds " \
                "with %.0f%% opened and " \
                "%i/%i mines marked. Press q " \
                "to quit or any key to play again.",
                time (0) - starttime,
                (percentage >= 99) ? 99 : percentage,
                marked, mines);
          status (msg);
          quit = true;
          timeout (-1);
          if (getch () == 'q')
            respawn = false;
        }
        break;
      case 'g':
        topline = 0;
        cury = 0;
        break;
      case 'G':
        if (lines >= LINES)
        {
          topline = lines-LINES+1;
          cury = LINES-2;
        }
        else
          cury = lines-1;
        break;
      case '^':
      case '0':
      case KEY_HOME:
        leftcol = 0;
        curx = 0;
        break;
      case '$':
      case KEY_END:
        if (cols > (COLS/2) + (COLS%2))
        {
          leftcol = cols*2 - COLS - COLS%2;
          curx = COLS-1-!(COLS%2);
        }
        else
          curx = COLS-1-!(COLS%2);
        break;
      case 'n':
        status ("Do you really want to start a new game? (y/N)");
        timeout (-1);
        if (getch () == 'y')
          main (argc, argv);
        else
          timeout (1000);
        break;
      case 'q':
        quit = true;
        respawn = false;
        continue;
      default:
        (input > 255) ?
          snprintf (msg, MAXMSG, "Unknown input %i", input):
          snprintf (msg, MAXMSG, "Unknown input %i", input);
          status (msg);
        continue;
    }

    percentage = (((float)opened/(float)((cols*lines)-mines))*100);

    if (opened == cols*lines - mines)
    {
      /* score: */
      ((time (0) - starttime) < 1) ?
        (percentage = 0):
        (percentage = ( (float)(cols*lines - whites - mines) /
         (float)(time (0) - starttime) ) /
         ( (float)(cols*lines) / (float)(mines) ) );
      snprintf (msg, MAXMSG,
          "%i mines swept in %li seconds, you win! " \
          "Your score is %.0f. " \
          "Press q to quit or any key to play again.",
          mines, time (0) - starttime, percentage*100);
      status (msg);
      timeout (-1);
      if (getch () == 'q')
        respawn = false;
      quit = true;
    }
    else if (!quit)
    {
      redraw (topline, leftcol, false);
      snprintf (msg, MAXMSG, "%i/%i: %.0f%% opened, %i/%i marked, " \
            "time %li seconds",
            cury+topline, (curx+leftcol)/2,
            (percentage >= 99) ? 99 : percentage,
            marked, mines,
            time (0) - starttime);
      status (msg);
    }
  }

  endwin ();
  free (field);

  if (respawn)
      main (argc, argv);

  return EXIT_SUCCESS;
}

void init (int argc, char *argv[])
{
  int i;

  initscr ();
  color = true;
  for (i=1; i<argc; i++)
  {
    if (argv[i][0] == '-')
      switch (argv[i][1])
      {
        case 'l':
          if (i+1 == argc)
          {
            endwin ();
            error (1, 0, "%s needs an argument", argv[i]);
          }
          if (!strcmp ("max", argv[++i]))
            lines = LINES-1;
          else
            lines = atoi (argv[i]);
          break;
        case 'c':
          if (i+1 == argc)
          {
            endwin ();
            error (1, 0, "%s needs an argument", argv[i]);
          }
          if (!strcmp ("max", argv[++i]))
            cols = (COLS/2) + (COLS%2);
          else
            cols = atoi (argv[i]);
          break;
        case 'm':
          if (i+1 == argc)
          {
            endwin ();
            error (1, 0, "%s needs an argument", argv[i]);
          }
          mines = atoi (argv[++i]);
          break;
        case 'n':
          color = false;
          break;
        case 'v':
          endwin ();
          version ();
          exit (EXIT_SUCCESS);
        case 'u':
          endwin ();
          usage (argv[0]);
        case 'x':
          if (i+1 == argc)
          {
            endwin ();
            error (1, 0, "%s needs an argument", argv[i]);
          }
          closed = *argv[++i];
          break;
        default:
          fprintf (stderr, "Unkown option %s\n", argv[i]);
      }
    else
      fprintf (stderr, "Unknown argument %s\n", argv[i]);
  }

  field = calloc (cols*lines, sizeof(short));
  if (!generate (field, cols, lines, mines))
    error (1, 0, "Generation failed because of invalid parameters");

  cury = (lines > LINES-1) ?
    ((LINES-1)/2):
    (lines/2);
  curx = (cols > (COLS/2) + (COLS%2)) ?
    ((COLS/2)/2):
    (cols-(cols%2));

  cbreak ();
  noecho ();
  keypad (stdscr, true);
  timeout (1000);

  if ( color && (color = has_colors()) )
  {
    start_color ();
    use_default_colors ();
    /* those are save anyway */
    init_pair (1, COLOR_BLUE, -1); /* blue */
    init_pair (3, COLOR_RED, -1); /* red */
    init_pair (9, -1, -1);
    init_pair (10, -1, -1);
    /* this is for those that aren't */
    //if (can_change_color ())
    if (init_color (COLOR_DARKGREEN, 0, 500, 0) != ERR)
    {
      //status ("Cool colors");
      //init_color (COLOR_DARKGREEN, 0, 500, 0);
      init_color (COLOR_DARKBLUE, 0, 0, 500);
      init_color (COLOR_DARKRED, 500, 0, 0);
      init_color (COLOR_DARKCYAN, 0, 500, 500);
      init_color (COLOR_DARKGREY, 250, 250, 250);
      init_color (COLOR_LIGHTGREY, 500, 500, 500);
      init_pair (2, COLOR_DARKGREEN, -1); /* dark green */
      init_pair (4, COLOR_DARKBLUE, -1); /* dark blue */
      init_pair (5, COLOR_DARKRED, -1); /* dark red */
      init_pair (6, COLOR_DARKCYAN, -1); /* dark cyan */
      init_pair (7, COLOR_DARKGREY, -1); /* black */
      init_pair (8, COLOR_LIGHTGREY, -1); /* dark grey */
    }
    else
    {
      //status ("Save colors");
      init_pair (2, COLOR_GREEN, -1); /* dark green */
      init_pair (4, COLOR_BLUE, -1); /* dark blue */
      init_pair (5, COLOR_RED, -1); /* dark red */
      init_pair (6, COLOR_CYAN, -1); /* dark cyan */
      init_pair (7, COLOR_WHITE, -1); /* black */
      init_pair (8, COLOR_WHITE, -1); /* dark grey */
    }
  }
}

void redraw (int upper, int left, bool open)
{
  char **linebuf, *ptr, debug[100];
  int i,j, current;

  linebuf = (char **)calloc (lines, sizeof(char *));
  for (i=0; i<lines; i++)
  {
    ptr = *(linebuf + i) = (char *)calloc ((2*cols)+1, sizeof(char));
    snprintf (debug, 100, "ptr is %p", ptr);
    //status (debug);

    for (j=0; j<cols; j++, ptr+=2)
    {
      current = *(field + j + cols*i);
      if (!current)
          sprintf (ptr, "  ");
      else if (open)
      {
        /* "totally" open */
        /*
        if ((current % 10) == 9)
            sprintf (ptr, "%c ", MINE_SYM);
        else if ((current % 10) == 0)
            sprintf (ptr, "  ");
        else
            sprintf (ptr, "%1i ", current%10);
        */

        /* "normal" open */
        if (current > 19)
          sprintf (ptr, "%c ", ERROR_SYM);
        else if (current == 19)
          sprintf (ptr, "%c ", MARK_SYM);
        else if (current > MINE)
          sprintf (ptr, "%c ", closed);
        else if (current == MINE)
          sprintf (ptr, "%c ", MINE_SYM);
        else
          sprintf (ptr, "%1i ", current);
      }
      else
      {
        if (current > 18)
          sprintf (ptr, "%c ", MARK_SYM);
        else if (current > 8)
          sprintf (ptr, "%c ", closed);
        else
          sprintf (ptr, "%1i ", current);
      }
    }
  }

  if (!color)
    for (i=upper; (i-upper)<LINES-1 && i < lines; i++)
    {
      *(linebuf[i]+left+COLS) = 0;
      mvaddstr ((i-upper), 0, left+linebuf[i]);
      free (linebuf[i]);
    }
  else
    for (i=upper; (i-upper)<LINES-1 && i < lines; i++)
    {
      move (i-upper, 0);
      ptr = left+linebuf[i];
      for (j=0; j<COLS && j<cols*2; j++, ptr++)
      {
        if (1 <= atoi (ptr) && atoi (ptr) <= 8)
          color_set (atoi (ptr), 0);
        else if (*ptr == closed
              || *ptr == MARK_SYM
              || *ptr == MINE_SYM
              || *ptr == ERROR_SYM)
          color_set (9, 0);

        if (*ptr == MINE_SYM
              || *ptr == MARK_SYM
              || *ptr == ERROR_SYM)
          addch (*ptr | A_BOLD);
        else
          addch (*ptr);

        //addch (' ');
      }
      free (linebuf[i]);
    }

  free (linebuf);
  move (cury, curx);

  refresh ();
}

void status (const char *string)
{
  if (color && color_set (10, 0) == ERR)
      return;
  mvaddstr (LINES-1, 0, string);
  clrtoeol ();
  move (cury, curx);
  refresh ();
}

int open (int x, int y)
{
  short *current = field + x + y*cols;
  int counter = 0;

  if (x < 0 || y < 0 || x >= cols || y >= lines
      || *current > 18)
    return 0;

  if (*current > MINE)
  {
    *current -= 10;
    counter++;
    if (*current == 0)
    {
      counter += open (x-1, y-1);
      counter += open (x, y-1);
      counter += open (x+1, y-1);
      counter += open (x-1, y);
      counter += open (x+1, y);
      counter += open (x-1, y+1);
      counter += open (x, y+1);
      counter += open (x+1, y+1);
    }

    return counter;
  }
  else if (*current == MINE)
    return -1;

  return 0;
}

int mark (int x, int y)
{
  short *victim = (field + x + y*cols);

  if (*victim < 9)
    return 0;
  else if (*victim < 18)
  {
    *victim += 10;
    return +1;
  }
  else
  {
    *victim -= 10;
    return -1;
  }
}

void usage (char *progname)
{
  printf ("Usage: %s [-c[olumns] n] [-l[ines] n] [-m[ines] n]" \
        " [-n[ocolor]] [-v[ersion]] [-u[sage]]\n", progname);
  puts ("Where:\n" \
        "  -c[olumns] n\tGenerate a map with n columns\n" \
        "  -l[ines] n\tGenerate a map with n lines\n" \
        "  -m[ines] n\tGenerate a map with n mines\n" \
        "  -n[ocolor]\tDo not use color (even if available)\n" \
        "  -v[ersion]\tPrint version information and exit\n" \
        "  -u[sage]\tPrint (this) usage message and exit\n");

  exit (EXIT_SUCCESS);
}

void version (void)
{
  printf ("fkmines version %i.%i\n\n", MAJOR, MINOR);
  puts ("This program comes with ABSOLUTELY NO WARRANTY.\n" \
        "This is free software, and you are welcome to\n" \
        "redistribute it under the conditions of the GNU\n" \
        "General Public License Version 3.\n");
}
